/* $LAAS: vimanWeb.c 2008/08/21 18:35:44 mallet $ */

/*
 * Copyright (c) 2000-2006,2008 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                       Anthony Mallet on Wed Aug  9 2000
 *                   adapted to viman by Xavier Broquere on Fri Sept 2011
 *
 */

/*
 * plugin arguments -----------------------------------------------------
 *
 * bank = <name>	image bank
 * image = <name>	image name
 * width = <int>	size of image
 * height = <int>
 * quality = <int 0-99>	jpeg quality
 * delay = <int>	delay (ms*10) between transmission
 */

#include <portLib.h>

#define _GNU_SOURCE
#include <dlfcn.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <taskLib.h>

#include <jpeglib.h>
#include <jerror.h>

#include <posterLib.h>
#include <h2timeLib.h>

#include <pomStruct.h>
#include <viamStruct.h>

#include "viam/viamlib.h"


extern void	viman_jpeg_mem_dest(j_compress_ptr cinfo,
			unsigned char *buffer, int size);
extern int	viman_jpeg_mem_size(j_compress_ptr cinfo);


/* --- local data ---------------------------------------------------- */

static int	vimanweb_compress(unsigned char *dst, unsigned char *src,
			int width, int height, int channels,
			int rowstride, int size, int quality);
static void	vimanweb_reduce(unsigned char *src, int ow, int oh,
			int rw, int rh, int channels);
static int	vimanweb_senderror(char *text, int n, FILE *f);


/* --- vimanPosterInit ------------------------------------------------ */

/** This function is called by picoweb, and I don't know exactly its
 * purpose */

int
vimanPosterInit()
{
  return OK;
}


/* --- webviman ------------------------------------------------------- */

void
webviman(FILE *f, int argc, char **argv, char **argn)
{
#define SEP	"jpgimage123imagejpg"

  POSTER_ID posterId;
  ViamImageBank *poster;
  ViamImageHeader *imageheader;
  int pWidth, pHeight, pChannels, pRowStride, pSize;

  H2TIME date, last;
  unsigned char *pixels, *compress;
  int i, length;
  STATUS status;
  char buf[512];
  struct timespec tv;

  char *bank;				/* plugin defaults */
  char *image;
  int width = 256;
  int height = 192;
  int quality = 30;
  int delay = -1;

  if (argc == 1) {
    void *m;
    void (*fptr)(FILE *, int, char **, char **);

    /* list posters */
/*     fprintf(stderr, "list posters\n"); */
/*     m = dlopen("vimanClient.so", RTLD_NOW); */
/*     fptr = dlsym(RTLD_NEXT, "webviman"); */
/*     fptr(f, argc, argv, argn); */
    return;
  }

  /* parse other args */
  for(i = 1; i < argc; i++) {
    if (argn[i] == NULL) continue;
    if (!strcmp(argn[i], "bank")) bank = argv[i];
    else if (!strcmp(argn[i], "image")) image = argv[i];
    else if (!strcmp(argn[i], "width")) width = atoi(argv[i]);
    else if (!strcmp(argn[i], "height")) height = atoi(argv[i]);
    else if (!strcmp(argn[i], "quality")) quality = atoi(argv[i]);
    else if (!strcmp(argn[i], "delay")) delay = atoi(argv[i]);
  }

  /* look for poster */
  sprintf(buf, "viman%s", bank);
  if (posterFind(buf, &posterId) != OK) {
    vimanweb_senderror("no poster", 404, f);
    return;
  }
  poster = (ViamImageBank *)posterAddr(posterId);

  if (posterTake(posterId, POSTER_READ) != OK) {
    vimanweb_senderror("no poster take", 404, f);
    return;
  }

  /* look for image */
  imageheader = NULL;
  for(i=0; i<poster->nImages; i++)
    if (!strcmp(poster->image[i].name.id, image)) {
      imageheader = &poster->image[i];
      break;
    }
  if (!imageheader) {
    vimanweb_senderror("no image", 404, f);
    return;
  }

  /* check size */
  if (width > imageheader->width) width = imageheader->width;
  if (height > imageheader->height) height = imageheader->height;

  pWidth = imageheader->width;
  pHeight = imageheader->height;
  pChannels = imageheader->nChannels;
  pRowStride = imageheader->widthStep;
  pSize = imageheader->imageSize;

  /* create internal buffer for image compression */
  pixels = (unsigned char*)malloc(pSize);
  if (pixels == NULL) {
    posterGive(posterId);
    vimanweb_senderror("no memory", 500, f);
    return;
  }

  compress =  (unsigned char*)malloc(pSize);
  if (compress == NULL) {
    posterGive(posterId);
    free(pixels);
    vimanweb_senderror("no memory", 500, f);
    return;
  }

  posterGive(posterId);

  /* start transmission */
  fprintf(f, "HTTP/1.1 200 OK\nPragma: no-cache\n");
  fprintf(f, "Server: picoweb 1.0\n");
  if (delay != -1) {
    fprintf(f, "Content-Type: multipart/x-mixed-replace;boundary=" SEP "\n\n");
    fprintf(f, "\n--" SEP "\n");
  } else {
    fprintf(f, "Connection: close\n");
  }
  fprintf(f, "Content-Type: image/jpeg\n");

  while(1) {
    if (delay != -1) {
      status = posterIoctl(posterId, FIO_GETDATE, (char *)&date);
      if (status == OK && date.ntick == last.ntick) {
	tv.tv_sec = delay / sysClkRateGet();
	tv.tv_nsec = (delay % sysClkRateGet()) * 1000000000/ sysClkRateGet();
	nanosleep(&tv, NULL);
	continue;
      }
      last = date;
    }

    if (posterTake(posterId, POSTER_READ) != OK) {
      vimanweb_senderror("viman(take)", 503, f); break;
    }
    if (imageheader->imageSize != pSize) {
      posterGive(posterId);
      vimanweb_senderror("viman(compress)", 503, f); break;
      fprintf(stderr, "image size changed\n");
      break;
    }
    memcpy(pixels,
	   &imageheader->data[imageheader->dataOffset], pSize);
    posterGive(posterId);

    if (pWidth != width || pHeight != height || pChannels == 3)
      /* reduce image and/or change BGR to RGB */
      vimanweb_reduce(pixels, pWidth, pHeight, width, height, pChannels);

    length = vimanweb_compress(compress, pixels, width, height,
			      pChannels, width*pChannels,
			      width*height*pChannels, quality);
    if (length < 0) {
      vimanweb_senderror("viman(compress)", 500, f); break;
    }

    fprintf(f, "Content-Length: %d\n\n", length);
    if (fwrite(compress, length, 1, f) != 1) {
      fprintf(stderr, "short write\n");
      break;
    }

    if (delay == -1) /* only send one image */ break;

    tv.tv_sec = delay / sysClkRateGet();
    tv.tv_nsec = (delay % sysClkRateGet()) * (1000000000/ sysClkRateGet());
    nanosleep(&tv, NULL);
    fprintf(f, "\n--" SEP "\nContent-Type: image/jpeg\n");
  }

  free(pixels);
  free(compress);
}


/* --- vimanweb_compress ---------------------------------------------- */

static int
vimanweb_compress(unsigned char *dst, unsigned char *src, int width,
		 int height, int channels, int rowstride, int size,
		 int quality)

{
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPROW row_pointer[1];
  int length;

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);

  cinfo.image_width = width;
  cinfo.image_height = height;
  cinfo.input_components = channels;
  cinfo.in_color_space = (channels == 3)?JCS_RGB:JCS_GRAYSCALE;

  viman_jpeg_mem_dest(&cinfo, dst, size);

  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, quality, FALSE);
  jpeg_start_compress(&cinfo, TRUE);

  while (cinfo.next_scanline < cinfo.image_height) {
    /* jpeg_write_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could pass more
     * than one scanline at a time if that's more convenient. */
    row_pointer[0] = src + cinfo.next_scanline * rowstride;
    jpeg_write_scanlines(&cinfo, row_pointer, 1);
  }

  jpeg_finish_compress(&cinfo);
  length = viman_jpeg_mem_size(&cinfo);
  jpeg_destroy_compress(&cinfo);
  return length;
}


/* --- vimanweb_reduce ------------------------------------------------ */

static void
vimanweb_reduce(unsigned char *src, int ow, int oh, int rw, int rh,
	       int channels)
{
  unsigned char *dst = src;
  double hrest, vrest, herr, verr;
  int hstep, vstep, htot;
  int i, j;

  hstep = ow/rw;			vstep = oh/rh;
  hrest = (double)ow/rw - hstep;	vrest = (double)oh/rh - vstep;

  /* one line is already skipped by column scan */
  vstep--;

  verr = 0.0;
  for(i=0;i<rh;i++) {
    herr = 0.0;
    htot = 0;
    for(j = 0; j < rw; j++) {
      switch(channels) {
	case 1:
	  *dst = *src;
	  dst++;
	  break;

	case 3: {
	  unsigned char s = src[0];

	  dst[0] = src[2]; /* XXX: change BGR to RGB ... */
	  dst[1] = src[1];
	  dst[2] = s;
	  dst+=3;
	  break;
	}
      }
      herr += hrest;
      if (herr >= 1.0) {
	herr -= 1.0;
	src += channels;
	htot++;
      }
      src += hstep * channels;
      htot += hstep;
    }
    src+= (ow-htot) * channels;

    verr += vrest;
    if (verr >= 1.0) {
      verr -= 1.0;
      src+=ow * channels;
    }
    src += ow*vstep * channels;
  }
}


/* --- vimanweb_senderror --------------------------------------------- */

static int
vimanweb_senderror(char *text, int n, FILE *f)
{
  char buf[64];

  switch(n) {
    case 404:
      strncpy(buf, "HTTP/1.0 404 Not Found\n\n", sizeof(buf));
      break;

    case 500:
      strncpy(buf, "HTTP/1.0 500 Internal Server Error\n\n", sizeof(buf));
      break;

    case 503:
      strncpy(buf, "HTTP/1.0 503 Service Unavailable\n\n", sizeof(buf));
      break;
  }

  fwrite(buf, 1, strlen(buf), f);
  fprintf(stderr, "picoweb: web report: %s %s\n", text, buf);
  return ERROR;
}
